#include <fstream>
#include <map>
#include "socket.h"
#include "resource.h"

const unsigned short default_port = 10051;
const int TASK_TRAY = (WM_USER + 1);

typedef gimite::socket_stream TCPSocket;


class Mutex
{
public:
	Mutex() { m_mutex = CreateMutex(NULL, FALSE, NULL); }
	~Mutex() { CloseHandle(m_mutex); }
	void lock() { WaitForSingleObject(m_mutex, INFINITE); }
	void unlock() { ReleaseMutex(m_mutex); }

private:
	HANDLE m_mutex;

	Mutex(const Mutex&); // Rs[֎~ 
	const Mutex& operator=(const Mutex&);
};

class ScopedLock
{
public:
	ScopedLock(Mutex& m) : m_mutex(m) { m_mutex.lock(); }
	~ScopedLock() { m_mutex.unlock(); }

private:
	Mutex& m_mutex;
};




class Thread
{
public:
	Thread() : thread_id(0), is_running(false), destruct(false) {}

	virtual ~Thread() {
		destruct = true;
		wait();
	}

	virtual void run() {
		is_running = true;
		CreateThread(NULL, 0, _exec, this, 0, &thread_id);
	}

	virtual void wait() {
		while(isRunning())
			Sleep(1);
	}

	virtual bool isRunning() {
		return is_running;
	}

protected:
	DWORD thread_id;
	volatile bool is_running;
	volatile bool destruct;

	virtual int exec() {
		onThreadExit();
		return 0;
	}

	virtual void onThreadExit() {
		is_running = false;
	}

	static DWORD WINAPI _exec(void *v) {
		Thread& t = *((Thread*)v);
		return t.exec();
	}

private:
	Thread(const Thread&); // Rs[֎~ 
	const Thread& operator=(const Thread&);
};






class Peer : public Thread
{
public:
	Peer(const std::string& h, unsigned short p=default_port) : host(h), port(p) {
	}

	bool send(const std::string& d) {
		if(!isRunning()) {
			data = d;
			run(); // \Pbg쐬s1b炢~̂ŃXbh 
			return true;
		}
		return false;
	}

	std::string host;
	unsigned short port;

private:
	std::string data;

	int exec() {
		TCPSocket sst(host, port);
		if(!!sst)
			sst.write(data.c_str(), data.size());

		onThreadExit();
		return 0;
	}
};


class Server : public Thread
{
public:
	~Server();

	void kill();

protected:
	bool setClipboardData(const std::string& dat);
	bool recieveClipboardData(TCPSocket& sst);
	int exec();
};

namespace {
	unsigned short port = default_port;
	HWND gwnd = 0;
	NOTIFYICONDATA nidNoftyIconData;
	Mutex mutex;
	Server server;

	typedef std::map<std::string, Peer*> peer_cont;
	peer_cont peers;


	void SetPort(unsigned short p) {
		server.kill();
		port = p;
		server.run();
	}

	bool ErasePeer(const std::string& address) {
		if(peers.find(address)!=peers.end()) {
			delete peers[address];
			peers.erase(address);
			return true;
		}
		return false;
	}

	bool InsertPeer(const std::string& address, unsigned short port=default_port) {
		if(peers.find(address)==peers.end()) {
			peers[address] = new Peer(address, port);
			return true;
		}
		return false;
	}

	bool LoadConfig() {
		std::fstream conf("tcb.conf", std::ios::in);
		if(!conf)
			return false;

		std::string l;
		while(std::getline(conf, l)) {
			char buf[128];
			std::string address;
			int res;
			int peer_port;
			if(sscanf(l.c_str(), "port: %d", &port)) {
			}
			else if(sscanf(l.c_str(), "peer: %[^:]:%d", buf, &peer_port)==2) {
				address = buf;
				if(InsertPeer(address, peer_port))
					printf("peer %s:%d added\n", address.c_str(), peer_port);
			}
			else if(sscanf(l.c_str(), "peer: %s\r", buf)==1) {
				address = buf;
				if(InsertPeer(address))
					printf("peer %s:%d added\n", address.c_str(), default_port);
			}
		}
		return true;
	}

	bool SaveConfig() {
		std::fstream conf("tcb.conf", std::ios::out);
		if(!conf)
			return false;
		conf << "port: " << port << std::endl;
		for(peer_cont::iterator p=peers.begin(); p!=peers.end(); ++p)
			conf << "peer: " << p->second->host << ":"<< p->second->port << std::endl;
		return true;
	}

	size_t SendClipboardData() {
		static std::string prev;
		std::string dat;
		{
			ScopedLock lock(mutex); // sync

			HGLOBAL hg;
			if(OpenClipboard(gwnd)) {
				if(hg = GetClipboardData(CF_TEXT)) {
					PTSTR strText = (PTSTR)GlobalLock(hg);
					dat = strText;
					GlobalUnlock(hg);
				}
				CloseClipboard();
			}
		}

		size_t sent = 0;
		if(!dat.empty() && dat!=prev) {
			char buf[64];
			sprintf(buf, "TCPCB: %d\r\n", dat.size());
			std::string sd = std::string(buf) + dat;

			for(peer_cont::iterator p=peers.begin(); p!=peers.end(); ++p) {
				if(p->second->send(sd))
					++sent;
			}

			prev = dat;
		}
		return sent;
	}
}



Server::~Server() {
	kill();
}

void Server::kill() {
	destruct = true;
	for(int i=0; i<10 && isRunning(); ++i) {
		TCPSocket sst("localhost", port); // ҂XbhEQv 
		Sleep(100);
	}
	destruct = false;
}

bool Server::setClipboardData(const std::string& dat) {
	ScopedLock lock(mutex); // sync

	if(OpenClipboard(gwnd)) {
		EmptyClipboard();

		HGLOBAL hg = GlobalAlloc(GHND | GMEM_SHARE , dat.size()+1);
		PTSTR mem = (PTSTR)GlobalLock(hg);
		lstrcpy(mem , dat.c_str());
		GlobalUnlock(hg);
		SetClipboardData(CF_TEXT , hg);
		CloseClipboard();
	}
	else {
		return false;
	}
	return true;
}

bool Server::recieveClipboardData(TCPSocket& sst) {
	static std::string prev;
	std::string dat;
	std::string l;
	size_t size = 0;
	bool res = false;

	std::getline(sst, l);
	if(sscanf(l.c_str(), "TCPCB: %d\r\n", &size)==1) {
		char buf[1024];
		size_t recved = 0;
		while(!!sst && recved<size) {
			size_t num = std::min<size_t>(1024, size-recved);
			sst.read(buf, num);
			dat.insert(dat.end(), buf, buf+num);
			recved+=num;
		}
		if(recved==size && dat!=prev) { // setClipboardDataVɃf[^𑗐M̂ŁAOƓf[^̓ubNB 
			setClipboardData(dat);
			prev = dat;
			res = true;
		}
	}
	return res;
}

int Server::exec() {
	gimite::server_stream_socket server(port);
	while(!!server && !destruct) {
		TCPSocket sst(server.accept());
		if(!destruct)
			recieveClipboardData(sst);
	}
	onThreadExit();
	return 0;
}











BOOL CALLBACK PortProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp) {
	switch (msg) {
	case WM_CLOSE:
		EndDialog(hwnd, IDOK);
		return TRUE;

	case WM_COMMAND:
		if(wp==IDC_BUTTON_INSERT) {
			char buf[128];
			unsigned int port;
			GetWindowText(GetDlgItem(hwnd, IDC_EDIT_PORT), buf, 128);
			if(sscanf(buf, "%d", &port)==1 && port<65536) {
				SetPort(port);
				PostMessage(hwnd, WM_CLOSE, 0, 0);
			}
		}
		else if(wp==IDC_BUTTON_CANCEL)
			PostMessage(hwnd, WM_CLOSE, 0, 0);
		return TRUE;
	}
	return FALSE;
}

BOOL CALLBACK PeerProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp) {
	switch (msg) {
	case WM_INITDIALOG:
		{
			char buf[32];
			sprintf(buf, "%d", port);
			SetWindowText(GetDlgItem(hwnd, IDC_EDIT_PORT), buf);
		}
		return TRUE;

	case WM_CLOSE:
		EndDialog(hwnd, IDOK);
		return TRUE;

	case WM_COMMAND:
		if(wp==IDC_BUTTON_INSERT) {
			char buf[128];
			char adr[128];
			unsigned int port;
			GetWindowText(GetDlgItem(hwnd, IDC_EDIT_ADDRESS), adr, 128);
			GetWindowText(GetDlgItem(hwnd, IDC_EDIT_PORT), buf, 128);
			if(sscanf(buf, "%d", &port)==1 && port<65536) {
				InsertPeer(adr, port);
				PostMessage(hwnd, WM_CLOSE, 0, 0);
			}
		}
		else if(wp==IDC_BUTTON_CANCEL)
			PostMessage(hwnd, WM_CLOSE, 0, 0);
		return TRUE;
	}
	return FALSE;
}


void RefreshList(HWND hwnd) {
	SendDlgItemMessage(hwnd, IDC_LIST_PEERS, LB_RESETCONTENT, 0, 0);

	char buf[256];
	for(peer_cont::iterator p=peers.begin(); p!=peers.end(); ++p) {
		Peer *r = p->second;
		sprintf(buf, "%s:%d", r->host.c_str(), r->port);
		SendDlgItemMessage(hwnd, IDC_LIST_PEERS, LB_ADDSTRING, 0, (LPARAM)buf);
	}
}
void RefreshPort(HWND hwnd) {
	char buf[256];
	sprintf(buf, "%d", port);
	SetWindowText(GetDlgItem(hwnd, IDC_STATIC_PORT), buf);
}

BOOL CALLBACK SettingProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp) {
	switch (msg) {
	case WM_INITDIALOG:
		RefreshList(hwnd);
		RefreshPort(hwnd);
		return TRUE;

	case WM_CLOSE:
		EndDialog(hwnd, IDOK);
		return TRUE;

	case WM_COMMAND:
		if(wp==IDC_BUTTON_PORT) {
			DialogBox((HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE), MAKEINTRESOURCE(IDD_PORT), hwnd, PortProc);
			RefreshPort(hwnd);
		}
		else if(wp==IDC_BUTTON_INSERT) {
			DialogBox((HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE), MAKEINTRESOURCE(IDD_PEER), hwnd, PeerProc);
			RefreshList(hwnd);
		}
		else if(wp==IDC_BUTTON_ERASE) {
			char buf[256];
			char adr[128];
			int port;
			SendDlgItemMessage(hwnd, IDC_LIST_PEERS, LB_GETTEXT,
				SendDlgItemMessage(hwnd, IDC_LIST_PEERS, LB_GETCURSEL, 0, 0), (LPARAM)buf);
			if(sscanf(buf, "%[^:]:%d", adr, &port)==2 && ErasePeer(adr))
				RefreshList(hwnd);
		}
		else if(wp==IDC_BUTTON_SAVE)
			SaveConfig();
		else if(wp==IDC_BUTTON_CANCEL)
			PostMessage(hwnd, WM_CLOSE, 0, 0);
		return TRUE;
	}
	return FALSE;
}

LRESULT CALLBACK WndProc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp) {
	static HWND hNextWnd;
	static HMENU rootmenu, popup;
	PTSTR strText;
	HDC hdc;
	PAINTSTRUCT ps;
	RECT rect;
	POINT point;

	switch (msg) {
	case WM_DESTROY:
		ChangeClipboardChain(hwnd , hNextWnd);
		Shell_NotifyIcon(NIM_DELETE, &nidNoftyIconData);
		PostQuitMessage(0);
		break;

	case WM_CREATE:
		rootmenu = LoadMenu(((LPCREATESTRUCT)(lp))->hInstance , MAKEINTRESOURCE(IDR_MENU));
		popup = GetSubMenu(rootmenu, 0);
		SetMenu(hwnd, rootmenu);
		hNextWnd = SetClipboardViewer(hwnd);
		break;

	case WM_DRAWCLIPBOARD:
		if(hNextWnd)
			SendMessage(hNextWnd , msg , wp , lp);
		SendClipboardData();
		InvalidateRect(hwnd , NULL , TRUE);
		break;

	case WM_CHANGECBCHAIN:
		if((HWND)wp == hNextWnd)
			hNextWnd = (HWND)lp;
		else if(hNextWnd)
			SendMessage(hNextWnd , msg , wp , lp);
		break;

	case WM_PAINT:
		hdc = BeginPaint(hwnd , &ps);
		{
			ScopedLock lock(mutex); // sync
			if(OpenClipboard(hwnd)) {
				HGLOBAL hg;
				if(hg = GetClipboardData(CF_TEXT)) {
					strText = (PTSTR)GlobalLock(hg);
					GetClientRect(hwnd , &rect);
					DrawText(hdc , strText , -1 , &rect , DT_LEFT);
					GlobalUnlock(hg);
				}
				CloseClipboard();
			}
		}
		EndPaint(hwnd , &ps);
		return 0;

	case WM_SIZE:
		switch(wp) {
		case SIZE_MINIMIZED:
			ShowWindow(hwnd, SW_HIDE);
			break;
		}
		break;

	case WM_COMMAND:
		if(wp==ID_MENU_EXIT)
			PostMessage(hwnd, WM_DESTROY, 0, 0);
		else if(wp==ID_MENU_SETTING)
			DialogBox((HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE), MAKEINTRESOURCE(IDD_SETTING), hwnd, SettingProc);
		break;

	case TASK_TRAY:
		switch(LOWORD(lp)) {
		case WM_LBUTTONDBLCLK:
			ShowWindow(hwnd, SW_SHOWNORMAL);
			SetForegroundWindow(hwnd);
			break;
		case WM_RBUTTONUP:
			GetCursorPos(&point);
			SetForegroundWindow(hwnd);
			TrackPopupMenu(popup , TPM_LEFTALIGN | TPM_BOTTOMALIGN , point.x, point.y, 0, hwnd, NULL);
			break;
		}
		break;
	}

	return DefWindowProc(hwnd , msg , wp , lp);
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR lpCmdLine, int nCmdShow) {
	gimite::startup_socket();

	HACCEL haccel;
	MSG msg;
	WNDCLASS winc;

//	freopen("stdout.txt", "wb", stdout);
	LoadConfig();

	winc.style			= CS_HREDRAW | CS_VREDRAW;
	winc.lpfnWndProc	= WndProc;
	winc.cbClsExtra		= 0;
	winc.cbWndExtra		= DLGWINDOWEXTRA;
	winc.hInstance		= hInstance;
	winc.hIcon			= LoadIcon(hInstance , TEXT("IDI_ICON1"));
	winc.hCursor		= LoadCursor(NULL , IDC_ARROW);
	winc.hbrBackground	= (HBRUSH)GetStockObject(WHITE_BRUSH);
	winc.lpszMenuName	= NULL;
	winc.lpszClassName	= TEXT("TCPCB");


	if(!RegisterClass(&winc))
		return -1;

	gwnd = CreateWindow(
		TEXT("TCPCB") , TEXT("tcp clipboard") ,
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT , CW_USEDEFAULT ,
		300, 200,
		NULL, NULL, hInstance, NULL
	);
	if(gwnd == NULL)
		return -1;


	nidNoftyIconData.cbSize = sizeof(NOTIFYICONDATA);
	nidNoftyIconData.hWnd = gwnd;
	nidNoftyIconData.uID = 1;
	nidNoftyIconData.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
	nidNoftyIconData.uCallbackMessage = TASK_TRAY;
	nidNoftyIconData.hIcon = LoadIcon(hInstance , TEXT("IDI_ICON1"));
	strcpy(nidNoftyIconData.szTip, "tcp clipboard");


	Shell_NotifyIcon(NIM_ADD, &nidNoftyIconData);
	ShowWindow(gwnd, nCmdShow);

	server.run();

	while(GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	for(peer_cont::iterator p=peers.begin(); p!=peers.end(); ++p)
		delete p->second;

	gimite::cleanup_socket();
	return msg.wParam;
}


